/****************************************************************/
/**@file CameraTools.cpp: 
 * Desctiption: Higher level abstraction above raw camera driver.*
 *              This code acts like driver, that provides 
 *              RT point with ball position			*
 * Dependency:							*
 ****************************************************************/
#include "hardware/OsSupp.h"

#include "images/raster.h"
#include "images/ras_comp.h"

#include "CameraTools.h"


const int CameraTool::DEFAULT_THR = 220; ///< Default threshold value


/** Read input points from camera object */
Point *CameraTool::GetInput(int Index)
{
switch(Index)
  {
  case 0: return &VideoX;      //motor 1 position
  case 1: return &VideoY;      //motor 2 position  
  case 2: return &WrThreshold; ///< Get threshold
  default: return NULL;
  }
};


/** Periodical execution recalculates ball positions. */
int CameraTool::Execute(void)
{
int posX, posY;
 
  if(Camera==NULL) return -1;

  try
    {
    Camera->GrabImage();		//GrabImage();               
    }
  catch(...)
    {	// problem with grab
    return -2;
    }

return 0;
}


/**
  This method formats camera raw image into required format. Moreover it
  also crops image to a rectangular shape. This makes signifficant speedup of whole
  application.
  @param[in] ImageBuf Buffer that receives image data in required format.
  @param[in] mode = 0 true color image; 1 true color as gray; 2 true color as mono
                     0x11 native gray 8bit;0x12 8 bit mono  */
bool CameraTool::ReadImageAs(std::vector<unsigned char> & ImageBuf, int mode)
{
int Width,Height;
char destplanes;
int y;
int LoopCount;
int OffsetX;

  if(Camera==NULL) return false;
  if(Camera->GetBufferData()==NULL) return false;
  
  iWidth=Width=Camera->GetWidth();
  iHeight=Height=Camera->GetHeight();
  if(iWidth<=0 || iHeight<=0) 
	return false;  

  OffsetX=0;

  if(mode & 0x80)
    {
    mode &= ~0x80;
    if(Width>Height)	//cut image to a rectangle
      {
      OffsetX = (Width-Height) / 2;
      iWidth = Height;
      }
    }

  destplanes = mode<0x10 ? 24 : 8;
  ImageBuf.resize((iWidth*iHeight*destplanes)/8);
  
  
  // recompute the image to monochrome according to mode
  unsigned char *i, *j;  
    

  for(y=0;y<Height;y++)
    {
    LoopCount=iWidth;
		//calculate buffer positions
    i = Camera->GetBufferData() + (OffsetX+y*Width)*Camera->GetDepth()/8;    

    switch(Camera->GetDepth())
      {			//RGB camera 24 bit
      case 24:
        j = &ImageBuf.front() + (Height-y-1)*iWidth*destplanes/8; //?? I do not understand why grey images are flipped??        
        
        switch(mode)
          {
	  case 0x0:	//converts from BGR to RGB
	    while(LoopCount-->0)
              {
	      j[0] = i[2];  //(i[1]+i[2])/2;
              j[1] = i[1];  //(i[0]+i[2])/2;
              j[2] = i[0];  //(i[0]+i[1])/2;
	      j+=3;
	      i+=3;	      
	      }
	    break;
  	  case 0x1:	//converts from BGR to RGB gray
	    while(LoopCount-->0)
              {
	      j[2]=j[1]=j[0] = (*(i++) + *(i++) + *(i++))/3;        
              j+=3;	      
	      }
            break;
          case 0x2:	//converts from BGR to RGB mono
            {
            int thr3 = 3*threshold;
	    while(LoopCount-->0)
              {
	      j[2]=j[1]=j[0] = (*(i++) + *(i++) + *(i++))>=thr3 ? 255 : 0;
              j+=3;	      
              }
            break;
            }
          case 0x11:	//converts from BGR to 8bit gray
	    while(LoopCount-->0)
              {
	      *(j++) = (*(i++) + *(i++) + *(i++))/3;
              }
            break;
          case 0x12:	//converts from BGR to 8bit mono
            {
            int thr3 = 3*threshold;
	    while(LoopCount-->0)
               {
	       *(j++) = (*(i++) + *(i++) + *(i++))>=thr3 ? 255 : 0;
               }
            break;
            }
          }
        break;


	//gray camera 8 bit
      case 8:      
        j = &ImageBuf.front() + (y)*iWidth*destplanes/8;
 
        switch(mode)
          {
          case 0:		//converts from gray to RGB (also gray)
	  case 0x1:	//converts from gray to RGB gray
	      while(LoopCount-->0)
                {
	        j[2]=j[1]=j[0]=*(i++);
	        j+=3;
	        }
          case 0x2:	//converts from gray to RGB mono
	      while(LoopCount-->0)
                {
	        j[2]=j[1]=j[0]= *(i++)>=threshold ? 255 : 0;
  	        j+=3;
	        }
          case 0x11:	//converts from gray to 8bit gray
	      while(LoopCount-->0)
                {
  	        *(j++) = *(i++);
                }
              break;
          case 0x12:	//converts from RGB to 8bit mono                    
  	      while(LoopCount-->0)
                 {
	         *(j++) = *(i++)>=threshold ? 255 : 0;
                 }
              break;            
          }
        break;        


      default:	     // //unsupported mode
        //wxMessageBox(_T("Unsupported image conversion!"), _T("ReadImageAs()"), wxOK| wxICON_EXCLAMATION, NULL);
	return false;      
      }
    }
    
  return true;
}


/**
  This method formats camera raw image into required format. Moreover it
  also crops image to a rectangular shape. This makes signifficant speedup of whole
  application.
  @param[in] ImageBuf Buffer that receives image data in required format.
  @param[in] mode = 0 true color image; 1 true color as gray; 2 true color as mono
                     0x11 native gray 8bit;0x12 8 bit mono  */
bool CameraTool::ReadImageAs(Raster2DAbstract *ImageBuf, int mode)
{
int Width,Height;
char destplanes;
int y;
int LoopCount;
int OffsetX;

  if(Camera==NULL) return false;
  if(Camera->GetBufferData()==NULL) return false;
  if(ImageBuf==NULL) return false;
  
  iWidth=Width=Camera->GetWidth();
  iHeight=Height=Camera->GetHeight();
  if(iWidth<=0 || iHeight<=0) 
	return false;  

  OffsetX=0;

  if(mode & 0x80)
    {
    mode &= ~0x80;
    if(Width>Height)	//cut image to a rectangle
      {
      OffsetX = (Width-Height) / 2;
      iWidth = Height;
      }
    }

  destplanes = mode<0x10 ? 24 : 8;
  if(destplanes!=ImageBuf->GetPlanes())
    return false;
  if(ImageBuf->Size1D!=iWidth || ImageBuf->Size2D!=iHeight)
    {
    ImageBuf->Allocate2D(iWidth,iHeight);
    }  
  
  // recompute the image to monochrome according to mode
  unsigned char *i, *j;  
    

  for(y=0;y<Height;y++)
    {
    LoopCount=iWidth;
		//calculate buffer positions
    i = Camera->GetBufferData() + (OffsetX+y*Width)*Camera->GetDepth()/8;    

    switch(Camera->GetDepth())
      {			//RGB camera 24 bit
      case 24:
        j = (unsigned char *)ImageBuf->GetRow(Height-y-1);		//?? I do not understand why grey images are flipped??        
        
        switch(mode)
          {
	  case 0x0:	//converts from BGR to RGB
	    while(LoopCount-->0)
              {
	      j[0] = i[2];  //(i[1]+i[2])/2;
              j[1] = i[1];  //(i[0]+i[2])/2;
              j[2] = i[0];  //(i[0]+i[1])/2;
	      j+=3;
	      i+=3;	      
	      }
	    break;
  	  case 0x1:	//converts from BGR to RGB gray
	    while(LoopCount-->0)
              {
	      j[2]=j[1]=j[0] = (*(i++) + *(i++) + *(i++))/3;        
              j+=3;	      
	      }
            break;
          case 0x2:	//converts from BGR to RGB mono
            {
            int thr3 = 3*threshold;
	    while(LoopCount-->0)
              {
	      j[2]=j[1]=j[0] = (*(i++) + *(i++) + *(i++))>=thr3 ? 255 : 0;
              j+=3;	      
              }
            break;
            }
          case 0x11:	//converts from BGR to 8bit gray
	    while(LoopCount-->0)
              {
	      *(j++) = (*(i++) + *(i++) + *(i++))/3;
              }
            break;
          case 0x12:	//converts from BGR to 8bit mono
            {
            int thr3 = 3*threshold;
	    while(LoopCount-->0)
               {
	       *(j++) = (*(i++) + *(i++) + *(i++))>=thr3 ? 255 : 0;
               }
            break;
            }
          }
        break;


	//gray camera 8 bit
      case 8:      
        j = (unsigned char *)ImageBuf->GetRow(y);
 
        switch(mode)
          {
          case 0:		//converts from gray to RGB (also gray)
	  case 0x1:	//converts from gray to RGB gray
	      while(LoopCount-->0)
                {
	        j[2]=j[1]=j[0]=*(i++);
	        j+=3;
	        }
          case 0x2:	//converts from gray to RGB mono
	      while(LoopCount-->0)
                {
	        j[2]=j[1]=j[0]= *(i++)>=threshold ? 255 : 0;
  	        j+=3;
	        }
          case 0x11:	//converts from gray to 8bit gray
	      while(LoopCount-->0)
                {
  	        *(j++) = *(i++);
                }
              break;
          case 0x12:	//converts from RGB to 8bit mono                    
  	      while(LoopCount-->0)
                 {
	         *(j++) = *(i++)>=threshold ? 255 : 0;
                 }
              break;            
          }
        break;        


      default:	     // //unsupported mode
        //wxMessageBox(_T("Unsupported image conversion!"), _T("ReadImageAs()"), wxOK| wxICON_EXCLAMATION, NULL);
	return false;      
      }
    }
    
  return true;
}

