/****************************************************************/
/**@file CritSecExt.h: 
 * Desctiption: Prototypes for extended critical section that   *
 *              supports RO and RW locks.                       *
 * Dependency: Windows (9x|2k|XP) 			        *
 ****************************************************************/
#ifndef __CRITICAL_SECTION_EXTENDED_H_
#define __CRITICAL_SECTION_EXTENDED_H_

#include "windows.h"


/**
 * Extended critical section.
 * Class provides read-only and read-write lock functions. 
 */
class CriticalSectionExt
{
	HANDLE no_access;
	mutable long rocounter;
	mutable CRITICAL_SECTION rwlock;

	void Init(void);

public:
	CriticalSectionExt(void) { Init(); }
	virtual ~CriticalSectionExt();

	void LockRO(void) const;
	void UnlockRO(void) const;
	void LockRW(void) const;
	void UnlockRW(void) const;
};



/**
 * Exception-safe read-only locking for CriticalSectionExt
 */
class CriticalSectionExtROLocker
{
  const CriticalSectionExt *cs;
public:
  explicit CriticalSectionExtROLocker(const CriticalSectionExt *c) :cs(c) { cs->LockRO(); }
  ~CriticalSectionExtROLocker() { cs->UnlockRO(); }
};

/**
 * Exception-safe read-write locking for CriticalSectionExt
 */
class CriticalSectionExtRWLocker
{
  const CriticalSectionExt * cs;
public:
  explicit CriticalSectionExtRWLocker(const CriticalSectionExt * c) :cs(c) { cs->LockRW(); }
  ~CriticalSectionExtRWLocker() { cs->UnlockRW(); }
};


/*******************************************************************************
 *
 * inline fuctions implementation
 *
 *******************************************************************************/

inline void CriticalSectionExt::Init()
{
	no_access = CreateEvent(NULL, TRUE, TRUE, NULL);
	InitializeCriticalSection(&rwlock);
	rocounter = 0L;
}

inline CriticalSectionExt::~CriticalSectionExt()
{
	EnterCriticalSection(&rwlock);
	LeaveCriticalSection(&rwlock);

	CloseHandle(no_access);
	DeleteCriticalSection(&rwlock);
}

inline void CriticalSectionExt::LockRO(void) const
{
	EnterCriticalSection(&rwlock);
	if (InterlockedIncrement(&rocounter) > 0)
		ResetEvent(no_access);
	LeaveCriticalSection(&rwlock);
}

inline void CriticalSectionExt::UnlockRO(void) const
{
	if (InterlockedDecrement(&rocounter) == 0)
		SetEvent(no_access);
}

inline void CriticalSectionExt::LockRW(void) const
{
	EnterCriticalSection(&rwlock);
	while (rocounter>0)		// the while cycle is for safety only as well as the timeout 10sec
		WaitForSingleObject(no_access, 10000);
}

inline void CriticalSectionExt::UnlockRW(void) const
{
	LeaveCriticalSection(&rwlock);
}

#endif // __CRITICAL_SECTION_EXTENDED_H_
